/*
 * Copyright (c) 2023 Gaia Sky - All rights reserved.
 *  This file is part of Gaia Sky, which is released under the Mozilla Public License 2.0.
 *  You may use, distribute and modify this code under the terms of MPL2.
 *  See the file LICENSE.md in the project root for full license details.
 */

/**
 * Common utilities to all nebulae shaders.
 */

// Noise buffer in the diffuseTexture channel.
#ifdef diffuseTextureFlag
uniform sampler2D u_diffuseTexture;
#define fetchDiffuse(uv, defaultValue) textureLod(u_diffuseTexture, (uv + 0.5) / 256.0, 0.0)
#else
#define fetchDiffuse(uv, defaultValue) defaultValue
#endif // diffuseTextureFlag

#ifndef GLSL_LIB_NEBULAE
#define GLSL_LIB_NEBULAE
#define PI 3.14159265
#define PI2 PI*2.0
#define R(p, a) p=cos(a)*p+sin(a)*vec2(p.y, -p.x)

float dot2(in vec2 v) {
    return dot(v, v);
}
float dot2(in vec3 v) {
    return dot(v, v);
}
float length2(vec2 p) {
    return sqrt(p.x * p.x + p.y * p.y);
}

float length8(vec2 p) {
    p = p * p;
    p = p * p;
    p = p * p;
    return pow(p.x + p.y, 1.0 / 8.0);
}
float hash13(vec3 p3) {
    p3 = fract(p3 * .1031);
    p3 += dot(p3, p3.zyx + 31.32);
    return fract((p3.x + p3.y) * p3.z);
}
// iq's noise
float noise(in vec3 x) {
    vec3 p = floor(x);
    vec3 f = fract(x);
    f = f * f * (3.0 - 2.0 * f);
    vec2 uv = (p.xy + vec2(37.0, 17.0) * p.z) + f.xy;
    vec2 rg = fetchDiffuse(uv, vec2(0.2, 1.0)).yx;
    return 1.0 - 0.92 * mix(rg.x, rg.y, f.z);
}
float fbm(vec3 p) {
    return noise(p * 0.06125) * 0.5 + noise(p * 0.125) * 0.25 + noise(p * 0.25) * 0.125;
}
vec3 desaturate(vec3 color, float amount) {
    vec3 gray = vec3(dot(vec3(0.2126, 0.7152, 0.0722), color));
    return vec3(mix(color, gray, amount));
}
vec4 get_quat_rotation(vec3 axis, float radians) {
    float d = 1.0 / length(axis);
    float l_ang;
    if (radians < 0.0) {
        l_ang = PI2 - (mod(-radians, PI2));
    } else {
        l_ang = mod(radians, PI2);
    }
    float l_sin = sin(l_ang / 2.0);
    float l_cos = cos(l_ang / 2.0);
    return normalize(vec4(d * axis.x * l_sin, d * axis.y * l_sin, d * axis.z * l_sin, l_cos));
}
// Rotate a vector given a quaternion.
vec3 rotate(vec3 v, vec4 q) {
    vec3 temp = cross(q.xyz, v) + q.w * v;
    return v + 2.0 * cross(q.xyz, temp);
}
// Computes the near and far distances of the intersections of the given ray
// with the sphere located at objPos with the given radius.
vec2 sphereIntersectionDistances(vec3 camPos, vec3 rayDir, vec3 objPos, float radius) {
    vec3 oc = camPos - objPos; // Vector from sphere center to camera position.
    float b = dot(oc, rayDir); // Half of the linear term in the quadratic.
    float c = dot(oc, oc) - radius * radius; // Constant term in the quadratic.
    float discriminant = b * b - c; // Discriminant of the quadratic equation.

    // Avoid numerical issues: treat very small negative discriminants as zero.
    if (discriminant < 0.0) {
        return vec2(-1.0, -1.0); // No intersection.
    }

    float sqrtDiscriminant = sqrt(max(discriminant, 0.0)); // Clamp to avoid NaN.
    float d0 = -b - sqrtDiscriminant; // Closest intersection point.
    float d1 = -b + sqrtDiscriminant; // Farthest intersection point.

    // If the camera is inside the sphere, c will be negative.
    // Ensure numerical stability: discard negative distances due to precision issues.
    d0 = max(length(camPos) - radius, 0.0);

    return vec2(d0, d1);
}
#endif
